import contextlib

__prog__ = "AminoExtract"
__version__ = "0.1.0"

# use contextlib to suppress the ImportError
# This may occur when this file is imported in setup.py as dependencies are not yet installed
with contextlib.suppress(ImportError):
    from AminoExtract.__main__ import get_feature_name_attribute, main



import gzip
import sys

import magic
import pandas as pd
from Bio import Seq, SeqIO

from AminoExtract.functions import log


# It reads in a GFF file and stores its contents as a GFFdataframe object
class GffDataFrame(object):
    def __init__(
        self, logger=log, inputfile: str | None = None, verbose: bool = False
    ) -> None:
        None if inputfile else sys.exit("Inputfile is not provided")
        if readable_file_type(inputfile):
            self.inputfile = inputfile
            self.log = logger
            self.verbose = verbose
            self.log.info(
                f"Parsing GFF input file: '[green]{inputfile}[/green]'"
            ) if verbose else None
            self._read()
            self._read_header()
        else:
            self.log = log
            self.verbose = verbose
            self.log.error(
                f"Input file is not readable: {inputfile}"
            ) if self.verbose else None
            sys.exit(1)

    def split_attributes_column(self) -> pd.DataFrame:
        """Takes a dataframe with a column called "attributes" that contains a string of attributes, and it
        returns a dataframe with the attributes split into separate columns

        Parameters
        ----------
        df : pd.DataFrame

        Returns
        -------
            A dataframe with the attributes column split into individual columns.

        """
        self.df["attributes"] = self.df["attributes"].apply(_attr_string_to_dict)
        df = self.df.join(pd.DataFrame(self.df["attributes"].to_dict()).T).drop(
            "attributes", axis=1
        )
        return df

    def _read(self) -> pd.DataFrame:
        if _is_gzipped(self.inputfile):
            return self._read_gff_gzipped()
        else:
            return self._read_gff_uncompressed()

    def _read_gff_gzipped(self) -> pd.DataFrame:
        self.df = pd.read_csv(
            self.inputfile,
            sep="\t",
            comment="#",
            names=[
                "seqid",
                "source",
                "type",
                "start",
                "end",
                "score",
                "strand",
                "phase",
                "attributes",
            ],
            compression="gzip",
        )
        self.df = self._split_attributes_column()
        return self.df

    def _read_gff_uncompressed(self) -> pd.DataFrame:
        self.df = pd.read_csv(
            self.inputfile,
            sep="\t",
            comment="#",
            names=[
                "seqid",
                "source",
                "type",
                "start",
                "end",
                "score",
                "strand",
                "phase",
                "attributes",
            ],
        )
        self.df = self._split_attributes_column()
        return self.df

    def _split_attributes_column(self) -> pd.DataFrame:
        """Takes a dataframe with a column called "attributes" that contains a string of attributes, and it
        returns a dataframe with the attributes split into separate columns

        Parameters
        ----------
        df : pd.DataFrame

        Returns
        -------
            A dataframe with the attributes column split into individual columns.

        """
        self.df["attributes"] = self.df["attributes"].apply(_attr_string_to_dict)
        columns = self.df.columns.tolist()
        # remove key-value pair from the dictionary in the attributes column if the key is already a column
        self.df["attributes"] = self.df["attributes"].apply(
            lambda attr: {k: v for k, v in attr.items() if k not in columns}
        )
        self.df = self.df.join(pd.DataFrame(self.df["attributes"].to_dict()).T).drop(
            "attributes", axis=1
        )
        return self.df

    def _read_header(self):
        self.header = ""
        if _is_gzipped(self.inputfile):
            with gzip.open(self.inputfile, "rt") as f:
                for line in f:
                    if line.startswith("#"):
                        self.header += line
                    else:
                        break
        else:
            with open(self.inputfile, "r") as f:
                for line in f:
                    if line.startswith("#"):
                        self.header += line
                    else:
                        break
        return self.header


def _attr_string_to_dict(string: str) -> dict:
    """Takes a string like "a=1;b=2;c=3" and returns a dictionary like {"a": "1", "b": "2", "c": "3"}

    Parameters
    ----------
    string : str
        The string to parse.

    Returns
    -------
        A dictionary with the key being the attribute name and the value being the attribute value.

    """
    return dict([x.split("=") for x in string.split(";") if "=" in x])


def _is_gzipped(infile: str) -> bool:
    """Returns `True` if the file is gzipped, and `False` otherwise

    Parameters
    ----------
    infile : str
        The path to the file to be checked.

    Returns
    -------
        The mime type of the file.

    """
    return magic.from_file(infile, mime=True) == "application/gzip"


def readable_file_type(infile: str) -> bool:
    """Returns `True` if the file is a plain text file or a gzip file, and `False` otherwise

    Parameters
    ----------
    infile : str
        The file to check.

    Returns
    -------
        A boolean value.

    """
    return bool(
        magic.from_file(infile, mime=True) == "text/plain" or "application/gzip"
    )


def read_gff(file: str, verbose: bool = True) -> GffDataFrame:
    """Reads a GFF file and returns a GFFdataframe object

    Parameters
    ----------
    file : str
        str
    verbose : bool, optional
        If True, print out the number of lines read in.

    Returns
    -------
        A GffDataFrame object.

    """
    return GffDataFrame(inputfile=file, verbose=verbose)


def read_fasta(file: str, verbose: bool = False) -> list:
    """Reads a FASTA file and returns a list of SeqRecord objects

    Parameters
    ----------
    file : str
        str

    Returns
    -------
        A list of SeqRecord objects

    """
    log.info(f"Parsing FASTA input file: '[green]{file}[/green]'") if verbose else None
    return list(SeqIO.parse(file, "fasta"))



import logging
import re as _re
import shutil
import textwrap
from argparse import SUPPRESS, ArgumentParser, RawTextHelpFormatter
from typing import IO, Optional

import rich
from rich.highlighter import NullHighlighter
from rich.logging import RichHandler

# Central logging object using Rich's logging library
FORMAT = "%(message)s"
logging.basicConfig(
    level="NOTSET",
    format=FORMAT,
    datefmt="[%x %X]",
    handlers=[
        RichHandler(
            show_path=False,
            omit_repeated_times=False,
            rich_tracebacks=True,
            markup=True,
            highlighter=NullHighlighter(),
        )
    ],
)
log = logging.getLogger("rich")


# A subclass of ArgumentParser.RawTextHelpFormatter that fixes spacing in the help text and respects bullet points.
# Especially useful for multi-line help texts combined with default values.
# This class is taken, though slightly refactored to better work for Python3, from the 'argparse-formatter' module from davesteele; https://github.com/davesteele/argparse_formatter
# Direct link to original code: https://github.com/davesteele/argparse_formatter/blob/a15d89a99e20b0cad4c389a2aa490c551cef4f9c/argparse_formatter/flexi_formatter.py
class FlexiFormatter(RawTextHelpFormatter):
    """
    Help message formatter which respects paragraphs and bulleted lists.
    Only the name of this class is considered a public API.
    """

    def _split_lines(self, text, width):
        return self._para_reformat(text, width)

    def _fill_text(self, text, width, indent):
        lines = self._para_reformat(text, width)
        return "\n".join(lines)

    def _indents(self, line):
        """Return line indent level and "sub_indent" for bullet list text."""

        indent = len(_re.match(r"( *)", line).group(1))
        if list_match := _re.match(r"( *)(([*\-+>]+|\w+\)|\w+\.) +)", line):
            sub_indent = indent + len(list_match.group(2))
        else:
            sub_indent = indent

        return (indent, sub_indent)

    def _split_paragraphs(self, text):
        """Split text in to paragraphs of like-indented lines."""

        text = textwrap.dedent(text).strip()
        text = _re.sub("\n\n[\n]+", "\n\n", text)

        last_sub_indent = None
        paragraphs = []
        for line in text.splitlines():
            (indent, sub_indent) = self._indents(line)
            is_text = _re.search(r"[^\s]", line) is not None

            if is_text and indent == sub_indent == last_sub_indent:
                paragraphs[-1] += f" {line}"
            else:
                paragraphs.append(line)

            last_sub_indent = sub_indent if is_text else None
        return paragraphs

    def _para_reformat(self, text, width):
        """Reformat text, by paragraph."""

        paragraphs = []
        for paragraph in self._split_paragraphs(text):

            (indent, sub_indent) = self._indents(paragraph)

            paragraph = self._whitespace_matcher.sub(" ", paragraph).strip()
            new_paragraphs = textwrap.wrap(
                text=paragraph,
                width=width,
                initial_indent=" " * indent,
                subsequent_indent=" " * sub_indent,
            )

            # Blank lines get eaten by textwrap, put it back with [' ']
            paragraphs.extend(new_paragraphs or [" "])

        return paragraphs


class QuickArgFormatter(FlexiFormatter):
    """
    A subclass of the `FlexiFormatter` subclass that adjusts the maximum width of the help text in relationship to the width of the terminal.
    Additionally, adds a default value to the help text if one is provided in the ArgParse constructor.
    """

    def __init__(self, prog):
        term_width = shutil.get_terminal_size().columns
        max_help_position = min(max(24, term_width // 2), 80)
        super().__init__(prog, max_help_position=max_help_position)

    def _get_help_string(self, action):
        """
        If the default value is not SUPPRESS and the word "default" is not in the help text, then add
        the default value to the help text
        Parameters
        ----------
        action
            The action object that is being processed.
        Returns
        -------
            The help text for the action.
        """
        help_text = action.help
        if (
            action.default != SUPPRESS
            and "default" not in help_text.lower()
            and action.default is not None
        ):
            help_text += f"\n ([underline]default: {str(action.default)}[/underline])"
        return help_text


class RichParser(ArgumentParser):
    """
    A subclass of `argparse.ArgumentParser` that overrides the `_print_message` method to use
    `rich.print` instead of `print`
    """

    def _print_message(self, message: str, file: Optional[IO[str]] = None) -> None:
        return rich.print(message)



import pathlib
import sys

from Bio import Seq

from AminoExtract.functions import log


def write_aa_file(
    AA_dict: dict[str, dict[str, Seq.Seq]],
    output: pathlib.Path,
    name: str,
    outtype: int,
) -> None:
    """Takes a dictionary of dictionaries of Seq.Seq objects, a pathlib.Path object, a string, and an
    integer, and writes the Seq.Seq objects to either a single file or multiple files

    Parameters
    ----------
    AA_dict : dict[str, dict[str, Seq.Seq]]
        a dictionary of dictionaries. The first dictionary is keyed by the sequence ID, and the second
    dictionary is keyed by the feature name. The value of the second dictionary is the amino acid
    sequence as a biopython Seq() object.
    output : pathlib.Path
        The path to the output file or directory as a (POSIX)Path object
    name : str
        The name of the output file.
    outtype : int
        0 = single file, 1 = multiple files

    """

    log.info(f"{'='*20} Writing output(s) {'='*20}")
    if outtype == 0:
        with open(output, "w") as out:
            for SeqID, features in AA_dict.items():
                for feature, aa in features.items():
                    log.info(
                        f"Writing '[cyan]{SeqID} - {feature}[/cyan]' to file '[green]{output}[/green]'"
                    )
                    out.write(f">{name}.{feature}\n{aa}\n")
    elif outtype == 1:
        if not output.exists():
            output.mkdir()
        for SeqID, features in AA_dict.items():
            for feature, aa in features.items():
                log.info(
                    f"Writing '[cyan]{SeqID} - {feature}[/cyan]' to file \"[green]{output / f'{name}_{feature}.faa'}[/green]\""
                )
                with open(output / f"{name}_{feature}.faa", "w") as out:
                    out.write(f">{SeqID}.{feature}\n{aa}\n")



import sys

from Bio.Seq import Seq
from rich import print

from AminoExtract.functions import log
from AminoExtract.reader import GffDataFrame


def Reverse_complement(seq: str) -> Seq:
    """Reverse complement a sequence

    Parameters
    ----------
    seq : str
        The sequence to reverse complement

    Returns
    -------
    Seq
        The reverse complement of the input sequence
    """
    seq_obj = Seq(seq)  # create a Seq object from the sequence
    return seq_obj.reverse_complement()  # type: ignore as BioPythons seq object is weird with typehints


def Extract_AminoAcids(
    GFFobj: GffDataFrame, SeqRecords: list, verbose: bool = False
) -> dict:
    """
    Extract amino acids from the SeqRecord objects based on the start and end positions of the GFFobj.df dataframe

    Parameters
    ----------
    GFFobj : GffDataFrame object
        GffDataFrame
    SeqRecords : list
        list of SeqRecord objects
    feature_type : str
        str

    Returns
    -------
    dict
        A dictionary with the SeqRecord id as the key and the amino acid sequences as the value.
    """

    log.info(
        "Extracting and translating the amino acid sequence(s) from the nucleotide sequence(s)"
    ) if verbose else None

    # create a dictionary with the SeqRecord.id as the key and the SeqRecord.seq as the value
    SeqDict = {record.id: record.seq for record in SeqRecords}
    # create an empty dictionary
    AA_dict = {record.id: {} for record in SeqRecords}

    # iterate through the dataframe
    for row in GFFobj.df.itertuples():
        try:
            name = row.Name
        except AttributeError:	
            log.warn("No '[green]Name[/green]' attribute found in GFF records. Using '[cyan]ID[/cyan]' instead") if verbose else None
            name = f"ID-{row.ID}"
        
        
        # get the sequence ID from the row
        seq_id = row.seqid

        # get the start and end positions from the row
        # subtract 1 from the start position to account for 0-based indexing
        # end position is also 0-based but we don't need to subtract 1 because of the way python slices function
        start, end = row.start - 1, row.end

        # get the nucleotide sequence from the sequence dictionary
        NucSequence = SeqDict[seq_id]

        # get the sequence slice from the start to the end position
        seq_slice = NucSequence[start:end].replace("-", "")

        # convert the sequence slice to a string
        seq_slice_str = str(seq_slice)

        # reverse complement the sequence slice if the strand is negative
        if row.strand == "-":
            seq_slice_str = Reverse_complement(seq_slice_str)

        # create a Seq object from the sequence slice
        seq_slice_obj = Seq(seq_slice_str)

        # translate the sequence slice to amino acids
        AASequence = seq_slice_obj.translate(to_stop=True)

        # add the amino acid sequence to the amino_acid_dict dictionary
        AA_dict[seq_id][name] = AASequence
    return AA_dict



import sys

import pandas as pd

from AminoExtract.functions import log
from AminoExtract.reader import GffDataFrame


def filter_name(frame: pd.DataFrame, Sequence_IDs: list) -> pd.DataFrame:
    """Takes a dataframe and a list of Biopythons SeqRecord objects, and returns a dataframe with only the rows that
    match the sequence IDs within the SeqRecord objects

    Parameters
    ----------
    frame : pd.DataFrame
        the dataframe you want to filter
    seq_ids : list
        a list of SeqRecord objects

    Returns
    -------
        A dataframe with the rows that have the seqids in the fasta_ids list.

    """
    return frame[frame["seqid"].isin(Sequence_IDs)]


def filter_feature_type(frame: pd.DataFrame, feature_type: str) -> pd.DataFrame:
    """Return a new dataframe containing only the rows of the input dataframe that have the specified
    feature type.

    Parameters
    ----------
    frame : pd.DataFrame
        the dataframe to filter
    feature_type : str
        The type of feature you want to filter for.

    Returns
    -------
        A dataframe with only the rows that containing the specified feature type.

    """
    # keep all rows if feature_type is "all"
    return frame if feature_type == "all" else frame[frame["type"] == feature_type]


def filter_gff(
    GffRecords: GffDataFrame, SeqRecords: list, feature_type: str, verbose: bool = False
) -> GffDataFrame:
    """Filter the GFF dataframe by feature type and sequence name

    Parameters
    ----------
    GffRecords : GffDataFrame object
        GffDataFrame
    SeqRecords : list
        list of SeqRecord objects
    feature_type : str
        str

    Returns
    -------
        A GffDataFrame object.

    """
    Sequence_IDs = [record.id for record in SeqRecords]
    log.info(
        f"Filtering GFF records to only contain the following information:\n * Feature type: '[green]{feature_type}[/green]'\n * Sequence IDs: '[green]{', '.join(Sequence_IDs)}[/green]'"
    ) if verbose else None
    GffRecords.df = filter_feature_type(
        filter_name(GffRecords.df, Sequence_IDs), feature_type
    )
    return GffRecords


def filter_sequences(gff: GffDataFrame, SeqRecords: list):
    """Takes a GffDataFrame object and a list of SeqRecord objects, and returns a list of SeqRecord objects
    that only contain the sequences that are specified in the GffDataFrame object.

    Parameters
    ----------
    gff : GffDataFrame
        The GffDataFrame object to filter
    SeqRecords : list
        A list of SeqRecord objects

    Returns
    -------
        A list of SeqRecord objects that only contain the sequences that are specified in the GffDataFrame object.

    """
    return [record for record in SeqRecords if record.id in gff.df["seqid"].tolist()]


def empty_dataframe(
    frame: pd.DataFrame = pd.DataFrame(), feature_type: str | None = None
) -> bool:
    if frame.empty or feature_type is None:
        log.warn(
            f"The GFF file is empty after filtering.\nThis might mean that there are no records within the GFF that match the sequence ID(s) in the given Fasta file.\nThis could also mean that there are no records within the GFF that match the feature type '[cyan]{feature_type}[/cyan]'.\nPlease check your inputs and try again."
        )
        return True
    return False



"""
Copyright (C) 2022  RIVM

https://github.com/RIVM-bioinformatics/AminoExtract
"""

import sys

from AminoExtract.args import validate_args
from AminoExtract.filter import empty_dataframe, filter_gff, filter_sequences
from AminoExtract.reader import read_fasta, read_gff
from AminoExtract.sequences import Extract_AminoAcids
from AminoExtract.writer import write_aa_file


def get_feature_name_attribute(
    input_gff: str, input_seq: str, feature_type: str
) -> dict[str, list[str]]:
    """This function takes a GFF file, a FASTA file, and a feature type, and returns a dictionary of the
        feature names for each sequence in the FASTA file

        Parameters
        ----------
        input_gff : str
            the path to the gff file
        input_seq : str
            The path to the fasta file containing the sequences you want to extract features from.
        feature_type : str
            This is the type of feature you want to extract from the GFF file. For example, if you want to
    extract the CDS features, you would enter "CDS". Options are "CDS", "gene" and "all".

        Returns
        -------
            A dictionary with the sequence id as the key and a list of the feature names as the value.

    """
    gff = read_gff(file=input_gff, verbose=False)
    seq = read_fasta(input_seq)
    gff_records = filter_gff(GffRecords=gff, SeqRecords=seq, feature_type=feature_type)
    FiltSequences = filter_sequences(gff, seq)
    SeqAttributes = {record.id: [] for record in FiltSequences}
    for row in gff_records.df.itertuples():
        SeqAttributes[row.seqid].append(row.Name)
    return SeqAttributes


def main() -> None:
    args = validate_args(sys.argv[1:])

    GFF_Obj = read_gff(file=args.features, verbose=True)
    fasta_records = read_fasta(args.input, verbose=True)

    GFF_Obj = filter_gff(
        GffRecords=GFF_Obj,
        SeqRecords=fasta_records,
        feature_type=args.feature_type,
        verbose=True,
    )
    sys.exit(1) if empty_dataframe(GFF_Obj.df, args.feature_type) else None
    SeqRecords = filter_sequences(GFF_Obj, fasta_records)

    AA_dict = Extract_AminoAcids(GFFobj=GFF_Obj, SeqRecords=SeqRecords, verbose=True)

    write_aa_file(AA_dict, args.output, args.name, args.outtype)



import sys

from AminoExtract import __prog__, __version__

if sys.version_info.major != 3 or sys.version_info.minor < 10:
    print("Error: you must execute setup.py using Python 3.10 or later")
    sys.exit(1)

from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name=__prog__,
    version=__version__,
    author="Florian Zwagemaker",
    project_urls={"Source Code": "https://github.com/RIVM-bioinformatics/AminoExtract"},
    license="MIT",
    author_email="ids-bioinformatics@rivm.nl",
    long_description=long_description,
    description="AminoExtract is an application to extract aminoacid sequences from a fasta file based on a GFF.",
    long_description_content_type="text/markdown",
    python_requires=">=3.10",
    packages=find_packages(),
    install_requires=["biopython", "pandas", "rich", "python-magic"],
    entry_points={
        "console_scripts": [
            "aminoextract = AminoExtract.__main__:main",
            "AminoExtract = AminoExtract.__main__:main",
        ]
    },
    keywords=[],
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Intended Audience :: Science/Research",
        "Operating System :: POSIX"]
)



import argparse
import os
import pathlib
import re
import sys

from rich import print

from AminoExtract import __prog__, __version__
from AminoExtract.functions import QuickArgFormatter, RichParser, log


def set_output_type(args: argparse.Namespace) -> argparse.Namespace:
    args.outtype = 0

    if not pathlib.Path(args.output).suffixes:
        log.info(
            f"The given output seems to be a directory.\nAll amino acid sequences will be written to individual files in this directory.\n([cyan]{args.output}[/cyan])"
        )
        args.outtype = 1
        return args
    log.info(
        f"The given output seems to be a file.\nAll amino acid sequences will be written to this file.\n([cyan]{args.output}[/cyan])"
    )
    return args


def check_features(args: argparse.Namespace) -> argparse.Namespace:
    if args.feature_type not in ["CDS", "gene", "all"]:
        log.error(
            f"[green]'{args.feature_type}'[/green] is not a valid feature type.\nPlease use any of the following: [bold]'CDS'[/bold],[bold]'gene'[/bold], or [bold]'all'[/bold]\nThese keywords are case-sensitive."
        )
        sys.exit(1)
    return args


def check_valid_output_filename(args: argparse.Namespace) -> argparse.Namespace:
    output_ext = pathlib.PurePath(args.output).name
    if not re.match("^[\w\-. ]+$", output_ext) or "/." in str(args.output):
        log.error(
            f"'[red]{output_ext}[/red]' does not seem to be a valid filename.\nPlease use only alphanumeric characters, underscores, and dashes."
        )
        sys.exit(1)
    return args


def check_file_ext(
    fname: str | None = None, choices: list[str] | None = None, ftype: str | None = None
) -> pathlib.Path | None:
    """Check if the file exists and has a valid extension

    Parameters
    ----------
    fname
        The name of the file to check.
    choices
        A list of valid file extensions for the file type.
    ftype
        This is the type of file you're checking. It's used in the error message.

    Returns
    -------
        The absolute path of the file.

    """
    if fname is not None and os.path.isfile(fname):
        file_exts = "".join(pathlib.Path(fname).suffixes)
        if choices is not None and file_exts not in choices:
            log.error(
                f"{fname} does not have a valid {ftype} file extension.\nPlease use any of the following extensions: [bold]{' '.join(choices)}[/bold]"
            )
            sys.exit(1)
        return pathlib.Path(fname).resolve()
    log.error(f"[green]'{fname}'[/green] is not a file.\n Exiting...")
    sys.exit(1)


def get_args(givenargs: list[str] | None = None) -> argparse.Namespace:
    """This function takes in a list of arguments, parses them, and returns a Namespace object

    Parameters
    ----------
    givenargs
        The arguments given to the script.

    Returns
    -------
        The arguments that are being passed to the program.

    """

    parser = RichParser(
        prog=f"[bold]{__prog__}[/bold]",
        usage=f"{__prog__} \[required options] \[optional options]",
        description=f"[bold underline]{__prog__}[/bold underline]: A quick tool to extract amino acid sequences from a FASTA file.",
        formatter_class=QuickArgFormatter,
        add_help=False,
    )

    req_args = parser.add_argument_group(
        title="[bold underline]Required Arguments[/bold underline]"
    )
    opt_args = parser.add_argument_group(
        "[bold underline]Optional Arguments[/bold underline]"
    )

    req_args.add_argument(
        "--input",
        "-i",
        type=lambda s: check_file_ext(s, [".fasta", ".fas", ".fna", ".fa"], "FASTA"),
        metavar="File",
        help="Input FASTA file with nucleotide sequences.",
        required=True,
    )

    req_args.add_argument(
        "--features",
        "-gff",
        metavar="File",
        type=lambda s: check_file_ext(s, [".gff", ".gff3"], "GFF"),
        help="GFF file containing the information of the amino acid sequences to extract.",
        required=True,
    )

    req_args.add_argument(
        "--output",
        "-o",
        type=lambda s: pathlib.Path(s).absolute(),
        metavar="Path",
        help="Output path, either a [underline cyan]file[/underline cyan] or [underline magenta]directory[/underline magenta].\n * If a file path is given, then all amino acid sequences will be written to this file.\n * If a directory path is given, then each amino acid sequence will be written to a separate file in this directory.\n [underline]Please see the docs for more info[/underline]",
        required=True,
    )

    req_args.add_argument(
        "--name",
        "-n",
        type=str,
        metavar="Text",
        help="Name of the sample that is being processed.\n * This will be used to create the fasta headers when all amino acid sequences are written to a single file.\n * If the output is going to be written to individual files in an output-directory then this name will be used as a prefix to create the output files.\n [underline]Please see the docs for more info[/underline]",
        required=True,
    )

    opt_args.add_argument(
        "--feature-type",
        "-ft",
        type=str,
        metavar="Text",
        default="CDS",
        help="Defines which feature types in the input gff will be processed to amino acid sequences. Defaults to 'CDS'.\nOptions are 'CDS', 'gene', and 'all'",
        required=False,
    )

    opt_args.add_argument(
        "--version",
        "-v",
        action="version",
        version=__version__,
        help=f"Print the {__prog__} version and exit.",
    )

    opt_args.add_argument(
        "--help",
        "-h",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit.",
    )

    # TODO: add extra arg for extra gff filters such as source, strand, etc.

    return parser.parse_args(givenargs)


def validate_args(givenargs: list[str]) -> argparse.Namespace:
    parsed_args = set_output_type(get_args(givenargs))
    parsed_args = check_features(parsed_args)
    return check_valid_output_filename(parsed_args)
